"""
``odak.raytracing``
===================
Provides necessary definitions for geometric optics. See "General Ray tracing procedure" from G.H. Spencerand M.V.R.K Murty for the theoratical explanation.

"""
# To get sub-modules.
from .ray import *
from .primitives import *
from .boundary import *
