"""
``odak.learn.wave``
===================
Provides necessary definitions for neural networks and learning algorithms. The definitions are based on torch framework.
"""
import odak.learn.wave
import odak.learn.tools
